import functools
from typing import Any, Dict, List, Optional
from opentelemetry.trace import Span

from .base import BaseInstrumentor
from quotientai.exceptions import logger


class QdrantInstrumentor(BaseInstrumentor):
    """
    OpenTelemetry instrumentor for Qdrant.
    Traces Qdrant operations including collection management, points operations, and queries.
    """
    
    def __init__(self, tracer_name: str = "quotientai.qdrant"):
        super().__init__(tracer_name)
        self._original_methods = {}
    
    def _instrument(self, **kwargs):
        """Instrument Qdrant classes and methods."""
        try:
            import qdrant_client
            self._instrument_qdrant_client(qdrant_client)
            logger.info("Qdrant instrumentation completed")
        except ImportError:
            logger.warning("Qdrant not installed, skipping instrumentation")
    
    def _uninstrument(self):
        """Uninstrument Qdrant classes and methods."""
        try:
            import qdrant_client
            self._restore_original_methods()
            logger.info("Qdrant uninstrumentation completed")
        except ImportError:
            logger.warning("Qdrant not installed, skipping uninstrumentation")
    
    def _instrument_qdrant_client(self, qdrant_client):
        """Instrument Qdrant client methods."""
        # Instrument QdrantClient class
        if hasattr(qdrant_client, 'QdrantClient'):
            client_class = qdrant_client.QdrantClient
            
            # Instrument collection management methods
            if hasattr(client_class, 'create_collection'):
                self._original_methods['client.create_collection'] = client_class.create_collection
                client_class.create_collection = self._wrap_create_collection(
                    client_class.create_collection
                )
            
            if hasattr(client_class, 'get_collections'):
                self._original_methods['client.get_collections'] = client_class.get_collections
                client_class.get_collections = self._wrap_get_collections(
                    client_class.get_collections
                )
            
            if hasattr(client_class, 'delete_collection'):
                self._original_methods['client.delete_collection'] = client_class.delete_collection
                client_class.delete_collection = self._wrap_delete_collection(
                    client_class.delete_collection
                )
            
            # Instrument points operations
            if hasattr(client_class, 'upsert'):
                self._original_methods['client.upsert'] = client_class.upsert
                client_class.upsert = self._wrap_upsert(client_class.upsert)
            
            if hasattr(client_class, 'search'):
                self._original_methods['client.search'] = client_class.search
                client_class.search = self._wrap_search(client_class.search)
            
            if hasattr(client_class, 'delete'):
                self._original_methods['client.delete'] = client_class.delete
                client_class.delete = self._wrap_delete(client_class.delete)
            
            if hasattr(client_class, 'scroll'):
                self._original_methods['client.scroll'] = client_class.scroll
                client_class.scroll = self._wrap_scroll(client_class.scroll)
            
            if hasattr(client_class, 'get'):
                self._original_methods['client.get'] = client_class.get
                client_class.get = self._wrap_get(client_class.get)
    
    def _wrap_create_collection(self, original_method):
        """Wrap create_collection method with tracing."""
        instrumentor = self  # Capture the instrumentor instance
        
        @functools.wraps(original_method)
        def wrapper(*args, **kwargs):
            # Extract self and other parameters
            if not args:
                raise TypeError("Missing 'self' argument")
            self_obj = args[0]
            other_args = args[1:]
            
            attributes = instrumentor._get_common_attributes("create_collection")
            attributes["db.system.name"] = "qdrant"
            
            # Extract parameters from kwargs or positional args
            collection_name = kwargs.get('collection_name')
            vectors_config = kwargs.get('vectors_config')
            
            # If not in kwargs, check positional args
            if not collection_name and other_args:
                collection_name = other_args[0]
            if not vectors_config and len(other_args) > 1:
                vectors_config = other_args[1]
            
            attributes["db.collection.name"] = collection_name
            
            if vectors_config:
                if hasattr(vectors_config, 'size'):
                    attributes["db.collection.dimension"] = vectors_config.size
                elif isinstance(vectors_config, dict) and 'size' in vectors_config:
                    attributes["db.collection.dimension"] = vectors_config['size']
            
            with instrumentor.tracer.start_as_current_span("qdrant.create_collection") as span:
                span.set_attributes(attributes)
                try:
                    result = original_method(*args, **kwargs)
                    span.set_attribute("db.operation.status", "completed")
                    return result
                except Exception as e:
                    span.set_attribute("db.operation.status", "error")
                    span.record_exception(e)
                    raise
        return wrapper
    
    def _wrap_get_collections(self, original_method):
        """Wrap get_collections method with tracing."""
        instrumentor = self  # Capture the instrumentor instance
        
        @functools.wraps(original_method)
        def wrapper(*args, **kwargs):
            attributes = instrumentor._get_common_attributes("list_collections")
            attributes["db.system.name"] = "qdrant"
            
            with instrumentor.tracer.start_as_current_span("qdrant.get_collections") as span:
                span.set_attributes(attributes)
                try:
                    result = original_method(*args, **kwargs)
                    span.set_attribute("db.operation.status", "completed")
                    if hasattr(result, 'collections'):
                        span.set_attribute("db.collections.count", len(result.collections))
                    return result
                except Exception as e:
                    span.set_attribute("db.operation.status", "error")
                    span.record_exception(e)
                    raise
        return wrapper
    
    def _wrap_delete_collection(self, original_method):
        """Wrap delete_collection method with tracing."""
        instrumentor = self  # Capture the instrumentor instance
        
        @functools.wraps(original_method)
        def wrapper(*args, **kwargs):
            # Extract self and other parameters
            if not args:
                raise TypeError("Missing 'self' argument")
            self_obj = args[0]
            other_args = args[1:]
            
            attributes = instrumentor._get_common_attributes("delete_collection")
            attributes["db.system.name"] = "qdrant"
            
            # Extract parameters from kwargs or positional args
            collection_name = kwargs.get('collection_name')
            
            # If not in kwargs, check positional args
            if not collection_name and other_args:
                collection_name = other_args[0]
            
            attributes["db.collection.name"] = collection_name
            
            with instrumentor.tracer.start_as_current_span("qdrant.delete_collection") as span:
                span.set_attributes(attributes)
                try:
                    result = original_method(*args, **kwargs)
                    span.set_attribute("db.operation.status", "completed")
                    return result
                except Exception as e:
                    span.set_attribute("db.operation.status", "error")
                    span.record_exception(e)
                    raise
        return wrapper
    
    def _wrap_upsert(self, original_method):
        """Wrap upsert method with tracing."""
        instrumentor = self  # Capture the instrumentor instance
        
        @functools.wraps(original_method)
        def wrapper(*args, **kwargs):
            # Extract self and other parameters
            if not args:
                raise TypeError("Missing 'self' argument")
            self_obj = args[0]
            other_args = args[1:]
            
            attributes = instrumentor._get_common_attributes("upsert")
            attributes["db.system.name"] = "qdrant"
            
            # Extract parameters from kwargs or positional args
            collection_name = kwargs.get('collection_name')
            points = kwargs.get('points')
            
            # If not in kwargs, check positional args
            if not collection_name and other_args:
                collection_name = other_args[0]
            if not points and len(other_args) > 1:
                points = other_args[1]
            
            attributes["db.collection.name"] = collection_name
            attributes["db.vector_count"] = len(points)
            attributes["db.ids_count"] = len(points)
            
            with instrumentor.tracer.start_as_current_span("qdrant.upsert") as span:
                span.set_attributes(attributes)
                try:
                    result = original_method(*args, **kwargs)
                    span.set_attribute("db.operation.status", "completed")
                    if hasattr(result, 'operation_id'):
                        span.set_attribute("db.operation.id", result.operation_id)
                    return result
                except Exception as e:
                    span.set_attribute("db.operation.status", "error")
                    span.record_exception(e)
                    raise
        return wrapper
    
    def _wrap_search(self, original_method):
        """Wrap search method with tracing."""
        instrumentor = self  # Capture the instrumentor instance
        
        @functools.wraps(original_method)
        def wrapper(*args, **kwargs):
            # Extract self and other parameters
            if not args:
                raise TypeError("Missing 'self' argument")
            self_obj = args[0]
            other_args = args[1:]
            
            attributes = instrumentor._get_common_attributes("query")
            attributes["db.system.name"] = "qdrant"
            
            # Extract parameters from kwargs or positional args
            collection_name = kwargs.get('collection_name')
            query_vector = kwargs.get('query_vector')
            query_filter = kwargs.get('query_filter')
            limit = kwargs.get('limit', 10)  # Default from Qdrant
            offset = kwargs.get('offset', 0)  # Default from Qdrant
            with_payload = kwargs.get('with_payload', True)  # Default from Qdrant
            with_vectors = kwargs.get('with_vectors', False)  # Default from Qdrant
            
            # If not in kwargs, check positional args
            if not collection_name and other_args:
                collection_name = other_args[0]
            if not query_vector and len(other_args) > 1:
                query_vector = other_args[1]
            if not query_filter and len(other_args) > 2:
                query_filter = other_args[2]
            if not limit and len(other_args) > 3:
                limit = other_args[3]
            if not offset and len(other_args) > 4:
                offset = other_args[4]
            if not with_payload and len(other_args) > 5:
                with_payload = other_args[5]
            if not with_vectors and len(other_args) > 6:
                with_vectors = other_args[6]
            
            attributes["db.collection.name"] = collection_name
            attributes["db.n_results"] = limit
            attributes["db.offset"] = offset
            if query_filter:
                attributes["db.filter"] = instrumentor._safe_json_dumps(query_filter)
            if query_vector:
                attributes["db.vector_count"] = 1
            
            with instrumentor.tracer.start_as_current_span("qdrant.search") as span:
                span.set_attributes(attributes)
                try:
                    result = original_method(*args, **kwargs)
                    span.set_attribute("db.operation.status", "completed")
                    
                    # Add retrieved documents if available
                    if hasattr(result, 'result') and result.result:
                        span.set_attribute("db.ids_count", len(result.result))
                        
                        # Format documents for span attributes
                        documents = []
                        for point in result.result:
                            doc = {"id": point.id, "score": point.score}
                            if hasattr(point, 'payload') and point.payload:
                                doc["metadata"] = point.payload
                            if hasattr(point, 'vector') and point.vector and with_vectors:
                                doc["content"] = str(point.vector[:10])  # Truncate for span
                            documents.append(doc)
                        
                        if documents:
                            span.set_attribute("db.query.retrieved_documents", instrumentor._format_documents_for_span(documents))
                    
                    return result
                except Exception as e:
                    span.set_attribute("db.operation.status", "error")
                    span.record_exception(e)
                    raise
        return wrapper
    
    def _wrap_delete(self, original_method):
        """Wrap delete method with tracing."""
        instrumentor = self  # Capture the instrumentor instance
        
        @functools.wraps(original_method)
        def wrapper(*args, **kwargs):
            # Extract self and other parameters
            if not args:
                raise TypeError("Missing 'self' argument")
            self_obj = args[0]
            other_args = args[1:]
            
            attributes = instrumentor._get_common_attributes("delete")
            attributes["db.system.name"] = "qdrant"
            
            # Extract parameters from kwargs or positional args
            collection_name = kwargs.get('collection_name')
            points_selector = kwargs.get('points_selector')
            
            # If not in kwargs, check positional args
            if not collection_name and other_args:
                collection_name = other_args[0]
            if not points_selector and len(other_args) > 1:
                points_selector = other_args[1]
            
            attributes["db.collection.name"] = collection_name
            
            # Handle different types of points_selector
            if hasattr(points_selector, 'points'):
                attributes["db.ids_count"] = len(points_selector.points)
            elif isinstance(points_selector, dict):
                if 'points' in points_selector:
                    attributes["db.ids_count"] = len(points_selector['points'])
                if 'filter' in points_selector:
                    attributes["db.filter"] = instrumentor._safe_json_dumps(points_selector['filter'])
            
            with instrumentor.tracer.start_as_current_span("qdrant.delete") as span:
                span.set_attributes(attributes)
                try:
                    result = original_method(*args, **kwargs)
                    span.set_attribute("db.operation.status", "completed")
                    if hasattr(result, 'operation_id'):
                        span.set_attribute("db.operation.id", result.operation_id)
                    return result
                except Exception as e:
                    span.set_attribute("db.operation.status", "error")
                    span.record_exception(e)
                    raise
        return wrapper
    
    def _wrap_scroll(self, original_method):
        """Wrap scroll method with tracing."""
        instrumentor = self  # Capture the instrumentor instance
        
        @functools.wraps(original_method)
        def wrapper(*args, **kwargs):
            # Extract self and other parameters
            if not args:
                raise TypeError("Missing 'self' argument")
            self_obj = args[0]
            other_args = args[1:]
            
            attributes = instrumentor._get_common_attributes("scroll")
            attributes["db.system.name"] = "qdrant"
            
            # Extract parameters from kwargs or positional args
            collection_name = kwargs.get('collection_name')
            scroll_filter = kwargs.get('scroll_filter')
            limit = kwargs.get('limit', 10)  # Default from Qdrant
            offset = kwargs.get('offset', 0)  # Default from Qdrant
            with_payload = kwargs.get('with_payload', True)  # Default from Qdrant
            with_vectors = kwargs.get('with_vectors', False)  # Default from Qdrant
            
            # If not in kwargs, check positional args
            if not collection_name and other_args:
                collection_name = other_args[0]
            if not scroll_filter and len(other_args) > 1:
                scroll_filter = other_args[1]
            if not limit and len(other_args) > 2:
                limit = other_args[2]
            if not offset and len(other_args) > 3:
                offset = other_args[3]
            if not with_payload and len(other_args) > 4:
                with_payload = other_args[4]
            if not with_vectors and len(other_args) > 5:
                with_vectors = other_args[5]
            
            attributes["db.collection.name"] = collection_name
            attributes["db.limit"] = limit
            attributes["db.offset"] = offset
            if scroll_filter:
                attributes["db.filter"] = instrumentor._safe_json_dumps(scroll_filter)
            
            with instrumentor.tracer.start_as_current_span("qdrant.scroll") as span:
                span.set_attributes(attributes)
                try:
                    result = original_method(*args, **kwargs)
                    span.set_attribute("db.operation.status", "completed")
                    
                    # Add retrieved documents if available
                    if hasattr(result, 'result') and result.result:
                        span.set_attribute("db.ids_count", len(result.result))
                        
                        # Format documents for span attributes
                        documents = []
                        for point in result.result:
                            doc = {"id": point.id}
                            if hasattr(point, 'payload') and point.payload:
                                doc["metadata"] = point.payload
                            if hasattr(point, 'vector') and point.vector and with_vectors:
                                doc["content"] = str(point.vector[:10])  # Truncate for span
                            documents.append(doc)
                        
                        if documents:
                            span.set_attribute("db.query.retrieved_documents", instrumentor._format_documents_for_span(documents))
                    
                    return result
                except Exception as e:
                    span.set_attribute("db.operation.status", "error")
                    span.record_exception(e)
                    raise
        return wrapper
    
    def _wrap_get(self, original_method):
        """Wrap get method with tracing."""
        instrumentor = self  # Capture the instrumentor instance
        
        @functools.wraps(original_method)
        def wrapper(*args, **kwargs):
            # Extract self and other parameters
            if not args:
                raise TypeError("Missing 'self' argument")
            self_obj = args[0]
            other_args = args[1:]
            
            attributes = instrumentor._get_common_attributes("get")
            attributes["db.system.name"] = "qdrant"
            
            # Extract parameters from kwargs or positional args
            collection_name = kwargs.get('collection_name')
            ids = kwargs.get('ids')
            with_payload = kwargs.get('with_payload', True)  # Default from Qdrant
            with_vectors = kwargs.get('with_vectors', False)  # Default from Qdrant
            
            # If not in kwargs, check positional args
            if not collection_name and other_args:
                collection_name = other_args[0]
            if not ids and len(other_args) > 1:
                ids = other_args[1]
            if not with_payload and len(other_args) > 2:
                with_payload = other_args[2]
            if not with_vectors and len(other_args) > 3:
                with_vectors = other_args[3]
            
            attributes["db.collection.name"] = collection_name
            attributes["db.ids_count"] = len(ids)
            
            with instrumentor.tracer.start_as_current_span("qdrant.get") as span:
                span.set_attributes(attributes)
                try:
                    result = original_method(*args, **kwargs)
                    span.set_attribute("db.operation.status", "completed")
                    
                    # Add retrieved documents if available
                    if hasattr(result, 'result') and result.result:
                        span.set_attribute("db.vector_count", len(result.result))
                        
                        # Format documents for span attributes
                        documents = []
                        for point in result.result:
                            doc = {"id": point.id}
                            if hasattr(point, 'payload') and point.payload:
                                doc["metadata"] = point.payload
                            if hasattr(point, 'vector') and point.vector and with_vectors:
                                doc["content"] = str(point.vector[:10])  # Truncate for span
                            documents.append(doc)
                        
                        if documents:
                            span.set_attribute("db.query.retrieved_documents", instrumentor._format_documents_for_span(documents))
                    
                    return result
                except Exception as e:
                    span.set_attribute("db.operation.status", "error")
                    span.record_exception(e)
                    raise
        return wrapper
    
    def _restore_original_methods(self):
        """Restore original methods."""
        try:
            import qdrant_client
            if hasattr(qdrant_client, 'QdrantClient'):
                client_class = qdrant_client.QdrantClient
                
                for method_name, original_method in self._original_methods.items():
                    if method_name == 'client.create_collection':
                        client_class.create_collection = original_method
                    elif method_name == 'client.get_collections':
                        client_class.get_collections = original_method
                    elif method_name == 'client.delete_collection':
                        client_class.delete_collection = original_method
                    elif method_name == 'client.upsert':
                        client_class.upsert = original_method
                    elif method_name == 'client.search':
                        client_class.search = original_method
                    elif method_name == 'client.delete':
                        client_class.delete = original_method
                    elif method_name == 'client.scroll':
                        client_class.scroll = original_method
                    elif method_name == 'client.get':
                        client_class.get = original_method
                
                self._original_methods.clear()
        except ImportError:
            pass 


from .chroma import ChromaInstrumentor
from .base import BaseInstrumentor
from .pinecone import PineconeInstrumentor  
from .qdrant import QdrantInstrumentor

__all__ = ["BaseInstrumentor", "ChromaInstrumentor", "PineconeInstrumentor", "QdrantInstrumentor"] 


from quotientai.resources.auth import AuthResource
from quotientai.resources.logs import LogsResource
from quotientai.resources.logs import AsyncLogsResource
from quotientai.tracing.core import TracingResource

__all__ = [
    "AuthResource",
    "LogsResource",
    "AsyncLogsResource",
    "TracingResource",
]



import typer

from rich.console import Console

from quotientai.client import QuotientAI
from quotientai.exceptions import QuotientAIError

console = Console()

app = typer.Typer(
    help="[green]Quotient CLI tool for managing artifacts.[/green]",
    rich_markup_mode="rich",
    no_args_is_help=True,
    pretty_exceptions_show_locals=False,
)
list_app = typer.Typer(
    help="Commands to list artifacts and resources",
    no_args_is_help=True,
)

app.add_typer(list_app, name="list")

MAX_LOGS = 10

@list_app.command(name="logs")
def list_logs(limit: int = MAX_LOGS):
    """Command to get all logs.
    
    Args:
        limit: Maximum number of logs to return (default: `MAX_LOGS`)
    """
    try:
        quotient = QuotientAI()
        response = quotient.logs.list(limit=limit)

        if len(response) > MAX_LOGS:  # Always show max 10 in CLI
            console.print(response[:MAX_LOGS])
            remaining = len(response) - MAX_LOGS
            console.print(f"[yellow]\n... {remaining} more logs available. Use --limit <number> or the SDK to view more.[/yellow]")
        else:
            console.print(response)
    except QuotientAIError as e:
        raise


if __name__ == "__main__": # pragma: no cover
    typer.run(app)



import functools
from typing import Any, Dict, List, Optional
from opentelemetry.trace import Span

from .base import BaseInstrumentor
from quotientai.exceptions import logger


class PineconeInstrumentor(BaseInstrumentor):
    """
    OpenTelemetry instrumentor for Pinecone.
    Traces Pinecone operations including index management, upserts, queries, and deletes.
    """
    
    def __init__(self, tracer_name: str = "quotientai.pinecone"):
        super().__init__(tracer_name)
        self._original_methods = {}
    
    def _instrument(self, **kwargs):
        """Instrument Pinecone classes and methods."""
        try:
            import pinecone
            self._instrument_pinecone_client(pinecone)
            logger.info("Pinecone instrumentation completed")
        except ImportError:
            logger.warning("Pinecone not installed, skipping instrumentation")
    
    def _uninstrument(self):
        """Uninstrument Pinecone classes and methods."""
        try:
            import pinecone
            self._restore_original_methods()
            logger.info("Pinecone uninstrumentation completed")
        except ImportError:
            logger.warning("Pinecone not installed, skipping uninstrumentation")
    
    def _instrument_pinecone_client(self, pinecone):
        """Instrument Pinecone client methods."""
        # Instrument Pinecone class
        if hasattr(pinecone, 'Pinecone'):
            pinecone_class = pinecone.Pinecone
            
            # Instrument create_index
            if hasattr(pinecone_class, 'create_index'):
                self._original_methods['pinecone.create_index'] = pinecone_class.create_index
                pinecone_class.create_index = self._wrap_create_index(
                    pinecone_class.create_index
                )
            
            # Instrument list_indexes
            if hasattr(pinecone_class, 'list_indexes'):
                self._original_methods['pinecone.list_indexes'] = pinecone_class.list_indexes
                pinecone_class.list_indexes = self._wrap_list_indexes(
                    pinecone_class.list_indexes
                )
            
            # Instrument delete_index
            if hasattr(pinecone_class, 'delete_index'):
                self._original_methods['pinecone.delete_index'] = pinecone_class.delete_index
                pinecone_class.delete_index = self._wrap_delete_index(
                    pinecone_class.delete_index
                )
            
            # Instrument Index class methods
            if hasattr(pinecone, 'Index'):
                index_class = pinecone.Index
                self._instrument_index_methods(index_class)
    
    def _instrument_index_methods(self, index_class):
        """Instrument Index class methods."""
        # Instrument upsert
        if hasattr(index_class, 'upsert'):
            self._original_methods['index.upsert'] = index_class.upsert
            index_class.upsert = self._wrap_upsert(index_class.upsert)
        
        # Instrument query
        if hasattr(index_class, 'query'):
            self._original_methods['index.query'] = index_class.query
            index_class.query = self._wrap_query(index_class.query)
        
        # Instrument delete
        if hasattr(index_class, 'delete'):
            self._original_methods['index.delete'] = index_class.delete
            index_class.delete = self._wrap_delete(index_class.delete)
        
        # Instrument fetch
        if hasattr(index_class, 'fetch'):
            self._original_methods['index.fetch'] = index_class.fetch
            index_class.fetch = self._wrap_fetch(index_class.fetch)
        
        # Instrument update
        if hasattr(index_class, 'update'):
            self._original_methods['index.update'] = index_class.update
            index_class.update = self._wrap_update(index_class.update)
    
    def _wrap_create_index(self, original_method):
        """Wrap create_index method with tracing."""
        instrumentor = self  # Capture the instrumentor instance
        
        @functools.wraps(original_method)
        def wrapper(self, *args, **kwargs):
            attributes = instrumentor._get_common_attributes("create_index")
            attributes["db.system.name"] = "pinecone"
            # Try to extract name, dimension, metric, spec from args/kwargs for attributes
            if len(args) > 0:
                attributes["db.index.name"] = args[0]
            elif "name" in kwargs:
                attributes["db.index.name"] = kwargs["name"]
            if len(args) > 1:
                attributes["db.index.dimension"] = args[1]
            elif "dimension" in kwargs:
                attributes["db.index.dimension"] = kwargs["dimension"]
            if len(args) > 2:
                attributes["db.create_index.metric"] = args[2]
            elif "metric" in kwargs:
                attributes["db.create_index.metric"] = kwargs["metric"]
            if "spec" in kwargs:
                attributes["db.create_index.spec"] = instrumentor._safe_json_dumps(kwargs["spec"])
            
            with instrumentor.tracer.start_as_current_span("pinecone.create_index") as span:
                span.set_attributes(attributes)
                try:
                    result = original_method(self, *args, **kwargs)
                    span.set_attribute("db.operation.status", "completed")
                    return result
                except Exception as e:
                    span.set_attribute("db.operation.status", "error")
                    span.record_exception(e)
                    raise
        return wrapper
    
    def _wrap_list_indexes(self, original_method):
        """Wrap list_indexes method with tracing."""
        instrumentor = self  # Capture the instrumentor instance
        
        @functools.wraps(original_method)
        def wrapper(self):
            attributes = instrumentor._get_common_attributes("list_indexes")
            attributes["db.system.name"] = "pinecone"
            
            with instrumentor.tracer.start_as_current_span("pinecone.list_indexes") as span:
                span.set_attributes(attributes)
                try:
                    result = original_method(self)
                    span.set_attribute("db.operation.status", "completed")
                    if hasattr(result, 'indexes'):
                        span.set_attribute("db.indexes.count", len(result.indexes))
                    return result
                except Exception as e:
                    span.set_attribute("db.operation.status", "error")
                    span.record_exception(e)
                    raise
        return wrapper
    
    def _wrap_delete_index(self, original_method):
        """Wrap delete_index method with tracing."""
        instrumentor = self  # Capture the instrumentor instance
        
        @functools.wraps(original_method)
        def wrapper(*args, **kwargs):
            # Extract self and other parameters
            if not args:
                raise TypeError("Missing 'self' argument")
            self_obj = args[0]
            other_args = args[1:]
            
            attributes = instrumentor._get_common_attributes("delete_index")
            attributes["db.system.name"] = "pinecone"
            
            # Extract parameters from kwargs or positional args
            name = kwargs.get('name')
            
            # If not in kwargs, check positional args
            if not name and other_args:
                name = other_args[0]
            
            attributes["db.index.name"] = name
            
            with instrumentor.tracer.start_as_current_span("pinecone.delete_index") as span:
                span.set_attributes(attributes)
                try:
                    result = original_method(*args, **kwargs)
                    span.set_attribute("db.operation.status", "completed")
                    return result
                except Exception as e:
                    span.set_attribute("db.operation.status", "error")
                    span.record_exception(e)
                    raise
        return wrapper
    
    def _wrap_upsert(self, original_method):
        """Wrap upsert method with tracing."""
        instrumentor = self  # Capture the instrumentor instance
        
        @functools.wraps(original_method)
        def wrapper(*args, **kwargs):
            # Extract self and other parameters
            if not args:
                raise TypeError("Missing 'self' argument")
            self_obj = args[0]
            other_args = args[1:]
            
            attributes = instrumentor._get_common_attributes("upsert")
            attributes["db.system.name"] = "pinecone"
            
            # Extract parameters from kwargs or positional args
            vectors = kwargs.get('vectors')
            namespace = kwargs.get('namespace')
            
            # If not in kwargs, check positional args
            if not vectors and other_args:
                vectors = other_args[0]
            if not namespace and len(other_args) > 1:
                namespace = other_args[1]
            
            if namespace:
                attributes["db.query.namespace"] = namespace
            
            # Count vectors and IDs
            if isinstance(vectors, list):
                attributes["db.vector_count"] = len(vectors)
                attributes["db.ids_count"] = len(vectors)
            elif isinstance(vectors, dict) and "vectors" in vectors:
                vectors_list = vectors["vectors"]
                attributes["db.vector_count"] = len(vectors_list)
                attributes["db.ids_count"] = len(vectors_list)
            
            with instrumentor.tracer.start_as_current_span("pinecone.index.upsert") as span:
                span.set_attributes(attributes)
                try:
                    result = original_method(*args, **kwargs)
                    span.set_attribute("db.operation.status", "completed")
                    if hasattr(result, 'upserted_count'):
                        span.set_attribute("db.upserted_count", result.upserted_count)
                    return result
                except Exception as e:
                    span.set_attribute("db.operation.status", "error")
                    span.record_exception(e)
                    raise
        return wrapper
    
    def _wrap_query(self, original_method):
        """Wrap query method with tracing."""
        instrumentor = self  # Capture the instrumentor instance
        
        @functools.wraps(original_method)
        def wrapper(*args, **kwargs):
            # Extract self and other parameters
            if not args:
                raise TypeError("Missing 'self' argument")
            self_obj = args[0]
            
            attributes = instrumentor._get_common_attributes("query")
            attributes["db.system.name"] = "pinecone"
            
            # Extract parameters from kwargs
            vector = kwargs.get('vector')
            id_param = kwargs.get('id')
            top_k = kwargs.get('top_k', 10)  # Default from Pinecone
            namespace = kwargs.get('namespace')
            filter_param = kwargs.get('filter')
            include_values = kwargs.get('include_values', False)
            include_metadata = kwargs.get('include_metadata', True)
            
            attributes["db.n_results"] = top_k
            if namespace:
                attributes["db.query.namespace"] = namespace
            if filter_param:
                attributes["db.filter"] = instrumentor._safe_json_dumps(filter_param)
            
            # Determine query type
            if vector is not None:
                attributes["db.vector_count"] = 1
                query_type = "vector"
            elif id_param is not None:
                query_type = "id"
            else:
                query_type = "unknown"
            
            attributes["db.query.type"] = query_type
            
            with instrumentor.tracer.start_as_current_span("pinecone.index.query") as span:
                span.set_attributes(attributes)
                try:
                    result = original_method(*args, **kwargs)
                    span.set_attribute("db.operation.status", "completed")
                    
                    # Add retrieved documents if available
                    if hasattr(result, 'matches') and result.matches:
                        span.set_attribute("db.ids_count", len(result.matches))
                        
                        # Format documents for span attributes
                        documents = []
                        for match in result.matches:
                            doc = {"id": match.id, "score": match.score}
                            if hasattr(match, 'metadata') and match.metadata:
                                doc["metadata"] = match.metadata
                            if hasattr(match, 'values') and match.values and include_values:
                                doc["content"] = str(match.values[:10])  # Truncate for span
                            documents.append(doc)
                        
                        if documents:
                            span.set_attribute("db.query.retrieved_documents", instrumentor._format_documents_for_span(documents))
                    
                    return result
                except Exception as e:
                    span.set_attribute("db.operation.status", "error")
                    span.record_exception(e)
                    raise
        return wrapper
    
    def _wrap_delete(self, original_method):
        """Wrap delete method with tracing."""
        instrumentor = self  # Capture the instrumentor instance
        
        @functools.wraps(original_method)
        def wrapper(*args, **kwargs):
            # Extract self and other parameters
            if not args:
                raise TypeError("Missing 'self' argument")
            self_obj = args[0]
            
            attributes = instrumentor._get_common_attributes("delete")
            attributes["db.system.name"] = "pinecone"
            
            # Extract parameters from kwargs
            ids = kwargs.get('ids')
            delete_all = kwargs.get('delete_all', False)
            namespace = kwargs.get('namespace')
            filter_param = kwargs.get('filter')
            
            attributes["db.delete_all"] = delete_all
            if namespace:
                attributes["db.query.namespace"] = namespace
            if filter_param:
                attributes["db.filter"] = instrumentor._safe_json_dumps(filter_param)
            if ids:
                attributes["db.ids_count"] = len(ids)
            
            with instrumentor.tracer.start_as_current_span("pinecone.index.delete") as span:
                span.set_attributes(attributes)
                try:
                    result = original_method(*args, **kwargs)
                    span.set_attribute("db.operation.status", "completed")
                    return result
                except Exception as e:
                    span.set_attribute("db.operation.status", "error")
                    span.record_exception(e)
                    raise
        return wrapper
    
    def _wrap_fetch(self, original_method):
        """Wrap fetch method with tracing."""
        instrumentor = self  # Capture the instrumentor instance
        
        @functools.wraps(original_method)
        def wrapper(*args, **kwargs):
            # Extract self and other parameters
            if not args:
                raise TypeError("Missing 'self' argument")
            self_obj = args[0]
            other_args = args[1:]
            
            attributes = instrumentor._get_common_attributes("fetch")
            attributes["db.system.name"] = "pinecone"
            
            # Extract parameters from kwargs or positional args
            ids = kwargs.get('ids')
            namespace = kwargs.get('namespace')
            
            # If not in kwargs, check positional args
            if not ids and other_args:
                ids = other_args[0]
            if not namespace and len(other_args) > 1:
                namespace = other_args[1]
            
            attributes["db.ids_count"] = len(ids)
            if namespace:
                attributes["db.query.namespace"] = namespace
            
            with instrumentor.tracer.start_as_current_span("pinecone.index.fetch") as span:
                span.set_attributes(attributes)
                try:
                    result = original_method(*args, **kwargs)
                    span.set_attribute("db.operation.status", "completed")
                    
                    # Add fetched documents if available
                    if hasattr(result, 'vectors') and result.vectors:
                        span.set_attribute("db.vector_count", len(result.vectors))
                        
                        # Format documents for span attributes
                        documents = []
                        for vector_id, vector_data in result.vectors.items():
                            doc = {"id": vector_id}
                            if hasattr(vector_data, 'metadata') and vector_data.metadata:
                                doc["metadata"] = vector_data.metadata
                            if hasattr(vector_data, 'values') and vector_data.values:
                                doc["content"] = str(vector_data.values[:10])  # Truncate for span
                            documents.append(doc)
                        
                        if documents:
                            span.set_attribute("db.query.retrieved_documents", instrumentor._format_documents_for_span(documents))
                    
                    return result
                except Exception as e:
                    span.set_attribute("db.operation.status", "error")
                    span.record_exception(e)
                    raise
        return wrapper
    
    def _wrap_update(self, original_method):
        """Wrap update method with tracing."""
        instrumentor = self  # Capture the instrumentor instance
        
        @functools.wraps(original_method)
        def wrapper(*args, **kwargs):
            # Extract self and other parameters
            if not args:
                raise TypeError("Missing 'self' argument")
            self_obj = args[0]
            other_args = args[1:]
            
            attributes = instrumentor._get_common_attributes("update")
            attributes["db.system.name"] = "pinecone"
            
            # Extract parameters from kwargs or positional args
            id_param = kwargs.get('id')
            values = kwargs.get('values')
            set_metadata = kwargs.get('set_metadata')
            namespace = kwargs.get('namespace')
            
            # If not in kwargs, check positional args
            if not id_param and other_args:
                id_param = other_args[0]
            if not values and len(other_args) > 1:
                values = other_args[1]
            if not set_metadata and len(other_args) > 2:
                set_metadata = other_args[2]
            if not namespace and len(other_args) > 3:
                namespace = other_args[3]
            
            attributes["db.update.id"] = id_param
            if namespace:
                attributes["db.query.namespace"] = namespace
            if set_metadata:
                attributes["db.update.metadata"] = instrumentor._safe_json_dumps(set_metadata)
            if values:
                attributes["db.vector_count"] = 1
            
            with instrumentor.tracer.start_as_current_span("pinecone.index.update") as span:
                span.set_attributes(attributes)
                try:
                    result = original_method(*args, **kwargs)
                    span.set_attribute("db.operation.status", "completed")
                    return result
                except Exception as e:
                    span.set_attribute("db.operation.status", "error")
                    span.record_exception(e)
                    raise
        return wrapper
    
    def _restore_original_methods(self):
        """Restore original methods."""
        try:
            import pinecone
            if hasattr(pinecone, 'Pinecone'):
                pinecone_class = pinecone.Pinecone
                
                for method_name, original_method in self._original_methods.items():
                    if method_name == 'pinecone.create_index':
                        pinecone_class.create_index = original_method
                    elif method_name == 'pinecone.list_indexes':
                        pinecone_class.list_indexes = original_method
                    elif method_name == 'pinecone.delete_index':
                        pinecone_class.delete_index = original_method
                
                if hasattr(pinecone, 'Index'):
                    index_class = pinecone.Index
                    if 'index.upsert' in self._original_methods:
                        index_class.upsert = self._original_methods['index.upsert']
                    if 'index.query' in self._original_methods:
                        index_class.query = self._original_methods['index.query']
                    if 'index.delete' in self._original_methods:
                        index_class.delete = self._original_methods['index.delete']
                    if 'index.fetch' in self._original_methods:
                        index_class.fetch = self._original_methods['index.fetch']
                    if 'index.update' in self._original_methods:
                        index_class.update = self._original_methods['index.update']
                
                self._original_methods.clear()
        except ImportError:
            pass 


import json
import jwt
import os
import random
import time
import traceback
import warnings

from pathlib import Path
from typing import Any, Dict, List, Optional, Union

import httpx

from quotientai import resources
from quotientai.exceptions import handle_errors, logger
from quotientai.resources.logs import LogDocument
from quotientai.tracing.core import TracingResource
from quotientai.types import DetectionType


class _BaseQuotientClient(httpx.Client):
    def __init__(self, api_key: str):
        try:
            token_dir = Path.home()
        except Exception:
            if Path("/root/").exists():
                token_dir = Path("/root")
            else:
                token_dir = Path.cwd()

        self.api_key = api_key
        self.token = None
        self.token_expiry = 0
        self.token_api_key = None
        self._token_path = (
            token_dir
            / ".quotient"
            / f"{api_key[-6:]+'_' if api_key else ''}auth_token.json"
        )

        # Try to load existing token
        self._load_token()

        # Set initial authorization header (token if valid, otherwise API key)
        auth_header = (
            f"Bearer {self.token}" if self._is_token_valid() else f"Bearer {api_key}"
        )

        super().__init__(
            base_url="https://api.quotientai.co/api/v1",
            headers={"Authorization": auth_header},
        )

    def _save_token(self, token: str, expiry: int):
        """Save token to memory and disk"""
        self.token = token
        self.token_expiry = expiry

        # Create directory if it doesn't exist
        try:
            self._token_path.parent.mkdir(parents=True, exist_ok=True)
        except Exception:
            logger.error(
                f"could not create directory for token. if you see this error please notify us at contact@quotientai.co"
            )
            return None

        # Save to disk
        with open(self._token_path, "w") as f:
            json.dump(
                {"token": token, "expires_at": expiry, "api_key": self.api_key}, f
            )

    def _load_token(self):
        """Load token from disk if available"""
        if not self._token_path.exists():
            return

        try:
            with open(self._token_path, "r") as f:
                data = json.load(f)
                self.token = data.get("token")
                self.token_expiry = data.get("expires_at", 0)
                self.token_api_key = data.get("api_key")
        except Exception:
            # If loading fails, token remains None
            pass

    def _is_token_valid(self):
        """Check if token exists and is not expired"""
        self._load_token()

        if not self.token:
            return False

        if self.token_api_key != self.api_key:
            return False

        # With 5-minute buffer
        return time.time() < (self.token_expiry - 300)

    def _update_auth_header(self):
        """Update authorization header with token or API key"""
        if self._is_token_valid():
            self.headers["Authorization"] = f"Bearer {self.token}"
        else:
            self.headers["Authorization"] = f"Bearer {self.api_key}"

    def _handle_response(self, response):
        """Check response for JWT token and save if present"""
        # Look for JWT token in response headers
        jwt_token = response.headers.get("X-JWT-Token")
        if jwt_token:
            try:
                # Parse token to get expiry (assuming token is a standard JWT)
                decoded = jwt.decode(jwt_token, options={"verify_signature": False})
                expiry = decoded.get("exp", time.time() + 3600)  # Default 1h if no exp

                # Save the token
                self._save_token(jwt_token, expiry)

                # Update auth header for future requests
                self.headers["Authorization"] = f"Bearer {jwt_token}"
            except Exception as e:
                # If token parsing fails, continue with current auth
                pass

        return response

    @handle_errors
    def _get(
        self, path: str, params: Optional[Dict[str, Any]] = None, timeout: int = None
    ) -> dict:
        """Send a GET request to the specified path."""
        self._update_auth_header()
        response = self.get(path, params=params, timeout=timeout)
        return self._handle_response(response)

    @handle_errors
    def _post(self, path: str, data: dict = {}, timeout: int = None) -> dict:
        """Send a POST request to the specified path."""
        self._update_auth_header()

        if isinstance(data, dict):
            data = {k: v for k, v in data.items() if v is not None}
        elif isinstance(data, list):
            data = [v for v in data if v is not None]

        response = self.post(
            url=path,
            json=data,
            timeout=timeout,
        )
        return self._handle_response(response)

    @handle_errors
    def _patch(self, path: str, data: dict = {}, timeout: int = None) -> dict:
        """Send a PATCH request to the specified path."""
        self._update_auth_header()

        data = {k: v for k, v in data.items() if v is not None}
        response = self.patch(
            url=path,
            json=data,
            timeout=timeout,
        )
        return self._handle_response(response)

    @handle_errors
    def _delete(self, path: str, timeout: int = None) -> dict:
        """Send a DELETE request to the specified path."""
        self._update_auth_header()

        response = self.delete(path, timeout=timeout)
        return self._handle_response(response)


class QuotientLogger:
    """
    Logger interface that wraps the underlying logs resource.
    This class handles both configuration (via init) and logging.
    """

    def __init__(self, logs_resource):
        self.logs_resource = logs_resource

        self.app_name: Optional[str] = None
        self.environment: Optional[str] = None
        self.tags: Dict[str, Any] = {}
        self.sample_rate: float = 1.0
        self.hallucination_detection: bool = False
        self.inconsistency_detection: bool = False
        self._configured = False
        self.hallucination_detection_sample_rate = 0.0

    def init(
        self,
        *,
        app_name: str,
        environment: str,
        tags: Optional[Dict[str, Any]] = {},
        sample_rate: float = 1.0,
        # New detection parameters (recommended)
        detections: Optional[List[DetectionType]] = None,
        detection_sample_rate: Optional[float] = None,
        # Deprecated detection parameters
        hallucination_detection: Optional[bool] = None,
        inconsistency_detection: Optional[bool] = None,
        hallucination_detection_sample_rate: Optional[float] = None,
    ) -> "QuotientLogger":
        """
        Configure the logger with the provided parameters and return self.
        This method must be called before using log().

        Args:
            app_name: The name of the application
            environment: The environment (e.g., "production", "development")
            tags: Optional tags to attach to all logs
            sample_rate: Sample rate for logging (0-1)

            # New detection parameters (recommended):
            detections: List of detection types to run by default
            detection_sample_rate: Sample rate for all detections (0-1)

            # Deprecated detection parameters:
            hallucination_detection: [DEPRECATED in 0.3.4] Use detections=[DetectionType.HALLUCINATION] instead
            inconsistency_detection: [DEPRECATED in 0.3.4] Use detections=[DetectionType.INCONSISTENCY] instead
            hallucination_detection_sample_rate: [DEPRECATED in 0.3.4] Use detection_sample_rate instead
        """
        if not app_name or not isinstance(app_name, str):
            logger.error("app_name must be a non-empty string")
            return None
        if not environment or not isinstance(environment, str):
            logger.error("environment must be a non-empty string")
            return None
        if not isinstance(tags, dict):
            logger.error("tags must be a dictionary")
            return None
        if not isinstance(sample_rate, float):
            logger.error("sample_rate must be a float")
            return None

        # Check for deprecated vs new detection parameter usage
        deprecated_detection_params_used = any(
            [
                hallucination_detection is not None,
                inconsistency_detection is not None,
                hallucination_detection_sample_rate is not None,
            ]
        )

        detection_params_used = any(
            [
                detections is not None,
                detection_sample_rate is not None,
            ]
        )

        # Prevent mixing deprecated and new detection parameters
        if deprecated_detection_params_used and detection_params_used:
            logger.error(
                "Cannot mix deprecated parameters (hallucination_detection, inconsistency_detection, hallucination_detection_sample_rate) "
                "with new detection parameters (detections, detection_sample_rate) in logger.init(). Please use new detection parameters."
            )
            return None

        # Handle deprecated parameters (with deprecation warnings)
        if deprecated_detection_params_used:
            warnings.warn(
                "Deprecated parameters (hallucination_detection, inconsistency_detection, hallucination_detection_sample_rate) "
                "are deprecated as of 0.3.4. Please use new detection parameters (detections, detection_sample_rate) instead.",
                DeprecationWarning,
                stacklevel=2,
            )

            # Convert deprecated to new format
            detections = []
            if hallucination_detection:
                detections.append(DetectionType.HALLUCINATION)
            # Note: inconsistency_detection is deprecated and not supported in v2

            detection_sample_rate = hallucination_detection_sample_rate or 0.0

        self.app_name = app_name
        self.environment = environment
        self.tags = tags or {}
        self.sample_rate = sample_rate

        if not (0.0 <= self.sample_rate <= 1.0):
            logger.error(f"sample_rate must be between 0.0 and 1.0")
            return None

        # Store detection configuration (converted to new format)
        self.detections = detections or []
        self.detection_sample_rate = detection_sample_rate or 0.0

        if not (0.0 <= self.detection_sample_rate <= 1.0):
            logger.error(f"detection_sample_rate must be between 0.0 and 1.0")
            return None

        # Keep old properties for backward compatibility with deprecated logger.log()
        self.hallucination_detection = DetectionType.HALLUCINATION in self.detections
        self.inconsistency_detection = (
            False  # inconsistency_detection is deprecated and not supported in v2
        )
        self.hallucination_detection_sample_rate = self.detection_sample_rate

        self._configured = True

        return self

    def _should_sample(self) -> bool:
        """
        Determine if the log should be sampled based on the sample rate. Intentionally naive client-side sampling.
        """
        return random.random() < self.sample_rate

    def log(
        self,
        *,
        user_query: str,
        model_output: str,
        documents: List[Union[str, LogDocument]] = None,
        message_history: Optional[List[Dict[str, Any]]] = None,
        instructions: Optional[List[str]] = None,
        tags: Optional[Dict[str, Any]] = {},
        hallucination_detection: Optional[bool] = None,
        inconsistency_detection: Optional[bool] = None,
    ):
        """
        Log the model interaction synchronously.

        Merges the default tags (set via init) with any runtime-supplied tags and calls the
        underlying non_blocking_create function.

        .. deprecated:: 0.3.1
            Use :meth:`quotient.log()` instead. This method will be removed in a future version.
        """
        warnings.warn(
            "quotient.logger.log() is deprecated as of 0.3.1 and will be removed in a future version. "
            "Please use quotient.log() instead.",
            DeprecationWarning,
            stacklevel=2,
        )

        if not self._configured:
            logger.error(
                f"Logger is not configured. Please call init() before logging."
            )
            return None

        # Merge default tags with any tags provided at log time.
        merged_tags = {**self.tags, **(tags or {})}

        # Use the instance variable as the default if not provided
        hallucination_detection = (
            hallucination_detection
            if hallucination_detection is not None
            else self.hallucination_detection
        )
        inconsistency_detection = (
            inconsistency_detection
            if inconsistency_detection is not None
            else self.inconsistency_detection
        )

        # Validate documents format
        if documents:
            for doc in documents:
                if isinstance(doc, str):
                    continue
                elif isinstance(doc, dict):
                    try:
                        LogDocument(**doc)
                    except Exception as e:
                        logger.error(
                            f"Invalid document format: Documents must include 'page_content' field and optional 'metadata' object with string keys."
                        )
                        return None
                else:
                    actual_type = type(doc).__name__
                    logger.error(
                        f"Invalid document type: Received {actual_type}, but documents must be strings or dictionaries."
                    )
                    return None

        if self._should_sample():
            # Convert to new parameters for resources layer
            detections = []
            if hallucination_detection:
                detections.append("hallucination")
            if inconsistency_detection:
                detections.append("inconsistency")

            log_id = self.logs_resource.create(
                app_name=self.app_name,
                environment=self.environment,
                detections=detections,
                detection_sample_rate=self.hallucination_detection_sample_rate,
                user_query=user_query,
                model_output=model_output,
                documents=documents,
                message_history=message_history,
                instructions=instructions,
                tags=merged_tags,
            )

            return log_id
        else:
            return None

    def poll_for_detection(
        self, log_id: str, timeout: int = 300, poll_interval: float = 2.0
    ):
        """
        Get Detection results for a log.

        This method polls the Detection endpoint until the results are ready or the timeout is reached.

        Args:
            log_id: The ID of the log to get Detection results for
            timeout: Maximum time to wait for results in seconds (default: 300s/5min)
            poll_interval: How often to poll the API in seconds (default: 2s)

        Returns:
            Log object with Detection results if successful, None otherwise

        .. deprecated:: 0.3.1
            Use :meth:`quotient.poll_for_detection()` instead. This method will be removed in a future version.
        """
        warnings.warn(
            "quotient.logger.poll_for_detection() is deprecated as of 0.3.1 and will be removed in a future version. "
            "Please use quotient.poll_for_detection() instead.",
            DeprecationWarning,
            stacklevel=2,
        )

        if not self._configured:
            logger.error(
                f"Logger is not configured. Please call init() before getting Detection results."
            )
            return None

        if not log_id:
            logger.error("Log ID is required for Detection")
            return None

        # Call the underlying resource method
        return self.logs_resource.poll_for_detection(log_id, timeout, poll_interval)


class QuotientTracer:
    """
    Tracer interface that wraps the underlying tracing resource.
    This class handles both configuration (via init) and tracing.
    """

    def __init__(self, tracing_resource: TracingResource):
        self.tracing_resource = tracing_resource
        self.app_name: Optional[str] = None
        self.environment: Optional[str] = None
        self.metadata: Optional[Dict[str, Any]] = None
        self.instruments: Optional[List[Any]] = None

        self._configured = False

    def init(
        self,
        *,
        app_name: str,
        environment: str,
        instruments: Optional[List[Any]] = None,
    ) -> "QuotientTracer":
        """
        Configure the tracer with the provided parameters and return self.
        This method must be called before using trace().
        """
        self.app_name = app_name
        self.environment = environment
        self.instruments = instruments

        # Configure the underlying tracing resource
        self.tracing_resource.configure(
            app_name=app_name,
            environment=environment,
            instruments=instruments,
        )

        self._configured = True

        return self

    def trace(self):
        """
        Decorator to trace function calls for Quotient.

        Example:
            quotient.tracer.init(app_name="my_app", environment="prod")

            @quotient.trace()
            def my_function():
                pass

            @quotient.trace()
            async def my_async_function():
                pass
        """
        if not self._configured:
            logger.error(
                f"tracer is not configured. Please call init() before tracing."
            )
            return lambda func: func

        # Call the tracing resource without parameters since it's now configured
        return self.tracing_resource.trace()

    def force_flush(self):
        """
        Force flush all pending spans to the collector.
        This is useful for debugging and ensuring spans are sent immediately.
        """
        self.tracing_resource.force_flush()


class QuotientAI:
    """
    A client that provides access to the QuotientAI API.

    The QuotientAI class provides synchronous methods to interact with the QuotientAI API,
    including logging data to Quotient and running detections.

    Args:
        api_key (Optional[str]): The API key to use for authentication. If not provided,
            will attempt to read from QUOTIENT_API_KEY environment variable.
    """

    def __init__(self, api_key: Optional[str] = None):
        self.api_key = api_key or os.environ.get("QUOTIENT_API_KEY")
        if not self.api_key:
            logger.error(
                "could not find API key. either pass api_key to QuotientAI() or "
                "set the QUOTIENT_API_KEY environment variable. "
                f"if you do not have an API key, you can create one at https://app.quotientai.co in your settings page"
            )

        _client = _BaseQuotientClient(self.api_key)
        self.auth = resources.AuthResource(_client)
        self.logs = resources.LogsResource(_client)
        self.tracing = resources.TracingResource(_client)

        # Create an unconfigured logger instance.
        self.logger = QuotientLogger(self.logs)
        self.tracer = QuotientTracer(self.tracing)

        try:
            self.auth.authenticate()
        except Exception as e:
            logger.error(
                "If you are seeing this error, please check that your API key is correct.\n"
                f"If the issue persists, please contact support@quotientai.co\n{traceback.format_exc()}"
            )
            return None

    def log(
        self,
        *,
        # New detection parameters (recommended)
        detections: Optional[List[DetectionType]] = None,
        detection_sample_rate: Optional[float] = None,
        # Deprecated detection parameters
        hallucination_detection: Optional[bool] = None,
        inconsistency_detection: Optional[bool] = None,
        hallucination_detection_sample_rate: Optional[float] = None,
        # Common input parameters
        user_query: Optional[str] = None,
        model_output: Optional[str] = None,
        documents: Optional[List[Union[str, LogDocument]]] = None,
        message_history: Optional[List[Dict[str, Any]]] = None,
        instructions: Optional[List[str]] = None,
        tags: Optional[Dict[str, Any]] = {},
    ):
        """
        Log the model interaction.

        Args:
            # New detection parameters (recommended):
            detections: List of detection types to run (replaces hallucination_detection/inconsistency_detection)
            detection_sample_rate: Sample rate for all detections 0-1 (replaces hallucination_detection_sample_rate)

            # Deprecated detection parameters:
            hallucination_detection: [DEPRECATED in 0.3.4] Use detections=[DetectionType.HALLUCINATION] instead
            inconsistency_detection: [DEPRECATED in 0.3.4] Use detections=[DetectionType.INCONSISTENCY] instead
            hallucination_detection_sample_rate: [DEPRECATED in 0.3.4] Use detection_sample_rate instead

            # Common input parameters:
            user_query: The user's input query
            model_output: The model's response
            documents: Optional list of documents (strings or LogDocument objects)
            message_history: Optional conversation history
            instructions: Optional list of instructions
            tags: Optional tags to attach to the log

        Returns:
            Log ID if successful, None otherwise
        """
        if not self.logger._configured:
            logger.error(
                "logger must be initialized with valid inputs before using log()."
            )
            return None

        # Check for deprecated vs new detection parameter usage
        deprecated_detection_params_used = any(
            [
                hallucination_detection is not None,
                inconsistency_detection is not None,
                hallucination_detection_sample_rate is not None,
            ]
        )

        detection_params_used = any(
            [detections is not None, detection_sample_rate is not None]
        )

        # Prevent mixing deprecated and new detection parameters
        if deprecated_detection_params_used and detection_params_used:
            logger.error(
                "Cannot mix deprecated parameters (hallucination_detection, inconsistency_detection, hallucination_detection_sample_rate) "
                "with new detection parameters (detections, detection_sample_rate). Please use new detection parameters."
            )
            return None

        # Handle deprecated parameters (with deprecation warnings)
        if deprecated_detection_params_used:
            warnings.warn(
                "Deprecated parameters (hallucination_detection, inconsistency_detection, hallucination_detection_sample_rate) "
                "are deprecated as of 0.3.4. Please use new detection parameters (detections, detection_sample_rate) instead. Document relevancy is not available with deprecated parameters.",
                DeprecationWarning,
                stacklevel=2,
            )

            # Convert deprecated to new format
            detections = []
            if hallucination_detection:
                detections.append(DetectionType.HALLUCINATION)

            detection_sample_rate = hallucination_detection_sample_rate or 0.0

            # For backward compatibility, require user_query and model_output
            if not user_query or not model_output:
                logger.error(
                    "user_query and model_output are required when using deprecated parameters"
                )
                return None

        # Handle new detection parameters
        else:
            # Use logger config as defaults if not provided in method call
            detections = (
                detections if detections is not None else self.logger.detections
            )
            detection_sample_rate = (
                detection_sample_rate
                if detection_sample_rate is not None
                else self.logger.detection_sample_rate
            )

            # Validate detection_sample_rate
            if not 0 <= detection_sample_rate <= 1:
                logger.error("detection_sample_rate must be between 0 and 1")
                return None

            # Validate required fields based on selected detections
            for detection in detections:
                if detection == DetectionType.HALLUCINATION:
                    if not user_query:
                        logger.error(
                            "user_query is required when hallucination detection is enabled"
                        )
                        return None
                    if not model_output:
                        logger.error(
                            "model_output is required when hallucination detection is enabled"
                        )
                        return None
                    if not documents and not message_history and not instructions:
                        logger.error(
                            "At least one of documents, message_history, or instructions must be provided when hallucination detection is enabled"
                        )
                        return None
                elif detection == DetectionType.DOCUMENT_RELEVANCY:
                    if not user_query:
                        logger.error(
                            "user_query is required when document_relevancy detection is enabled"
                        )
                        return None
                    if not documents:
                        logger.error(
                            "documents must be provided when document_relevancy detection is enabled"
                        )
                        return None

        # Convert DetectionType enums to strings for the resources layer
        detection_strings = [detection.value for detection in detections]

        log_id = self.logs.create(
            app_name=self.logger.app_name,
            environment=self.logger.environment,
            detections=detection_strings,
            detection_sample_rate=detection_sample_rate,
            user_query=user_query,
            model_output=model_output,
            documents=documents,
            message_history=message_history,
            instructions=instructions,
            tags={**(self.logger.tags or {}), **(tags or {})},
        )
        return log_id

    def trace(self):
        """Direct access to the tracer's trace decorator."""
        return self.tracer.trace()

    def poll_for_detection(
        self, log_id: str, timeout: int = 300, poll_interval: float = 2.0
    ):
        """
        Direct access to the logger's poll_for_detection method.

        Args:
            log_id: The ID of the log to get Detection results for
            timeout: Maximum time to wait for results in seconds (default: 300s/5min)
            poll_interval: How often to poll the API in seconds (default: 2s)
        """
        if not self.logger._configured:
            logger.error(
                "Logger is not configured. Please call quotient.logger.init() before using poll_for_detection()."
            )
            return None

        if not log_id:
            logger.error("Log ID is required for Detection")
            return None

        detection = self.logger.poll_for_detection(
            log_id=log_id, timeout=timeout, poll_interval=poll_interval
        )
        return detection

    def force_flush(self):
        """
        Force flush all pending spans to the collector.
        This is useful for debugging and ensuring spans are sent immediately.
        """
        self.tracer.force_flush()



from .client import QuotientAI
from .types import DetectionType
from .async_client import AsyncQuotientAI
from .exceptions import QuotientAIError

__all__ = ["QuotientAI", "QuotientAIError", "AsyncQuotientAI", "DetectionType"]
